/* $Id: smokeToML.cc,v 1.10 2004/12/09 12:42:38 yurug Exp $ */

#include <kde/smoke.h>

extern "C" {
# include <stdio.h>
# include <caml/memory.h>
# include <caml/mlvalues.h>
# include <caml/alloc.h>
# include <caml/custom.h>
# include <caml/intext.h>
# include <caml/fail.h>
# include <caml/callback.h>
# include <caml/bigarray.h>
# include <assert.h>
  value nikoret__classes_size (value binding);
  value nikoret__methods_size (value binding);
  value nikoret__types_size (value binding);
  value nikoret__class_name (value binding, value idx);
  value nikoret__method_name (value binding, value idx);
  value nikoret__method_flags (value binding, value idx);
  value nikoret__class_flags (value binding, value idx);
  value nikoret__method_return_type_index (value binding, value idx);
  value nikoret__class_of_method (value binding, value idx);
  value nikoret__method_arguments_index (value binding, value idx);
  value nikoret__extract_group (int idx, Smoke::Index* groups);
  value nikoret__inheritance_group (value binding, value idx);
  value nikoret__argument_group (value binding, value idx);
  value nikoret__ambiguous_method_group (value binding, value idx);
  value nikoret__class_parents (value binding, value idx);
  value nikoret__type_name (value binding, value idx);
  value nikoret__type_id (value binding, value idx);
  value nikoret__type_class (value binding, value idx);
}

typedef struct Smoke* c_smoke_ptr_t;

c_smoke_ptr_t cast_to_c (value binding)
{
  assert (Tag_val (binding) == Custom_tag);
  return (c_smoke_ptr_t) (Field (binding, 1));
}

value nikoret__classes_size (value binding)
{
  CAMLparam1 (binding);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  
  CAMLreturn (Val_int (sbinding->numClasses));
}

value nikoret__methods_size (value binding)
{
  CAMLparam1 (binding);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
    
  CAMLreturn (Val_int (sbinding->numMethods));
}

value nikoret__types_size (value binding)
{
  CAMLparam1 (binding);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
    
  CAMLreturn (Val_int (sbinding->numTypes));
}

static const char* noname = "NoName";

value nikoret__class_name (value binding, value idx)
{
  CAMLparam2 (binding, idx);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  const char* name = sbinding->classes[Int_val (idx)].className;
  if (name == NULL)
    name = noname;
  CAMLreturn (copy_string (name));
}

value nikoret__method_name (value binding, value idx)
{
  CAMLparam2 (binding, idx);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  Smoke::Index name_idx = sbinding->methods[Int_val (idx)].name;
  const char* name = sbinding->methodNames[name_idx];
  if (name == NULL)
    name = noname;
  CAMLreturn (copy_string (name));
}

value nikoret__method_return_type_index (value binding, value idx)
{
  CAMLparam2 (binding, idx);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  CAMLreturn (Val_int (sbinding->methods[Int_val (idx)].ret));
}

value push_method_flag (value list, char* tag, Smoke::MethodFlags f, int flags)
{
  CAMLparam0 ();
  CAMLlocal1 (r);
  
  if (flags & f) {
    r = alloc (2, 0);
    Store_field (r, 0, hash_variant (tag));
    Store_field (list, 1, r);
  }
  else r = list;
  CAMLreturn (r);
}

value nikoret__method_flags (value binding, value idx)
{
  CAMLparam2 (binding, idx);
  CAMLlocal1 (t);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  int flags = sbinding->methods[Int_val (idx)].flags;
  t = alloc (2, 0);
  value head = t;
  t = push_method_flag (t, "MFStatic"         , Smoke::mf_static  , flags);
  t = push_method_flag (t, "MFConst"          , Smoke::mf_const   , flags);
  t = push_method_flag (t, "MFCopyConstructor", Smoke::mf_copyctor, flags);
  t = push_method_flag (t, "MFInternal"       , Smoke::mf_internal, flags);
  Store_field (t, 1, Val_int (0));
  CAMLreturn (head);
}

value push_class_flag (value list, char* tag, Smoke::ClassFlags f, int flags)
{
  CAMLparam0 ();
  CAMLlocal1 (r);
  
  if (flags & f) {
    r = alloc (2, 0);
    Store_field (r, 0, hash_variant (tag));
    Store_field (list, 1, r);
  }
  else r = list;
  CAMLreturn (r);
}

value nikoret__class_flags (value binding, value idx)
{
  CAMLparam2 (binding, idx);
  CAMLlocal1 (t);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  int flags = sbinding->classes[Int_val (idx)].flags;
  t = alloc (2, 0);
  value head = t;
  t = push_class_flag (t, "CFConstructor"    , Smoke::cf_constructor, flags);
  t = push_class_flag (t, "CFDeepCopy"       , Smoke::cf_deepcopy   , flags);
  t = push_class_flag (t, "CFVirtual"        , Smoke::cf_virtual    , flags);
  t = push_class_flag (t, "CFUndefined"      , Smoke::cf_undefined  , flags);
  Store_field (t, 1, Val_int (0));
  CAMLreturn (head);  
}

value nikoret__class_of_method (value binding, value idx)
{
  CAMLparam2 (binding, idx);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  CAMLreturn (Val_int (sbinding->methods[Int_val (idx)].classId));
}

value nikoret__method_arguments_index (value binding, value idx)
{
  CAMLparam2 (binding, idx);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  CAMLreturn (Val_int (sbinding->methods[Int_val (idx)].args));
}

value nikoret__extract_group (int idx, Smoke::Index* groups)
{
  long dims[1];
  int c = idx;
  while (groups[c] != 0)
    ++c;
  dims[0] = c - idx;
  return alloc_bigarray(BIGARRAY_SINT16 | BIGARRAY_C_LAYOUT,
			1, 
			groups + idx, 
			dims);  
}

value nikoret__inheritance_group (value binding, value idx)
{
  CAMLparam2 (binding, idx);

  c_smoke_ptr_t sbinding = cast_to_c (binding);
  CAMLreturn (nikoret__extract_group (Int_val (idx), 
				      sbinding->inheritanceList));
}

value nikoret__argument_group (value binding, value idx)
{
  CAMLparam2 (binding, idx);

  c_smoke_ptr_t sbinding = cast_to_c (binding);
  CAMLreturn (nikoret__extract_group (Int_val (idx), 
				      sbinding->argumentList));
}

value nikoret__ambiguous_method_group (value binding, value idx)
{
  CAMLparam2 (binding, idx);

  c_smoke_ptr_t sbinding = cast_to_c (binding);
  CAMLreturn (nikoret__extract_group (Int_val (idx), 
				      sbinding->ambiguousMethodList));
}

value nikoret__class_parents (value binding, value idx)
{
  CAMLparam2 (binding, idx);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  int parents = sbinding->classes[Int_val (idx)].parents;
  CAMLreturn (Val_int (parents));
}

value nikoret__type_name (value binding, value idx)
{
  CAMLparam2 (binding, idx);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  const char* name = sbinding->types[Int_val (idx)].name;
  if (name == NULL)
    name = noname;
  CAMLreturn (copy_string (name));
}

value nikoret__type_class (value binding, value idx)
{
  CAMLparam2 (binding, idx);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  CAMLreturn (Val_int (sbinding->types[Int_val (idx)].classId));
}

value nikoret__type_id (value binding, value idx)
{
  value modifiers = alloc_tuple (3);
  c_smoke_ptr_t sbinding = cast_to_c (binding);
  unsigned short flags = sbinding->types[Int_val (idx)].flags;

  switch (flags & Smoke::tf_elem) {
  case Smoke::t_voidp:
    Store_field (modifiers, 0, hash_variant ("TIVoidP")); break;
  case Smoke::t_bool:
    Store_field (modifiers, 0, hash_variant ("TIBool")) ; break;
  case Smoke::t_char:
    Store_field (modifiers, 0, hash_variant ("TIChar")) ; break;
  case Smoke::t_uchar:
    Store_field (modifiers, 0, hash_variant ("TIUChar")); break;
  case Smoke::t_short:
    Store_field (modifiers, 0, hash_variant ("TIShort")); break;
  case Smoke::t_ushort:
    Store_field (modifiers, 0, hash_variant ("TIUShort"));break;
  case Smoke::t_int:
    Store_field (modifiers, 0, hash_variant ("TIInt"))  ; break;
  case Smoke::t_uint:
    Store_field (modifiers, 0, hash_variant ("TIUInt")) ; break;
  case Smoke::t_long:
    Store_field (modifiers, 0, hash_variant ("TILong")) ; break;
  case Smoke::t_ulong:
    Store_field (modifiers, 0, hash_variant ("TIULong")); break;
  case Smoke::t_float:
    Store_field (modifiers, 0, hash_variant ("TIFloat")); break;
  case Smoke::t_double:
    Store_field (modifiers, 0, hash_variant ("TIDouble"));break;
  case Smoke::t_enum:
    Store_field (modifiers, 0, hash_variant ("TIEnum")) ; break;
  case Smoke::t_class:
    Store_field (modifiers, 0, hash_variant ("TIClass")); break;
  case Smoke::t_last:
    Store_field (modifiers, 0, hash_variant ("TILast")) ; break;
  }

  flags &= ~Smoke::tf_elem;

  if ((flags & Smoke::tf_const) != 0)
    Store_field (modifiers, 2, Val_int (1));
  else 
    Store_field (modifiers, 2, Val_int (1));
  flags &= ~Smoke::tf_const;

  if (flags == Smoke::tf_stack)
    Store_field (modifiers, 1, hash_variant ("TFStack"));
  else if (flags == Smoke::tf_ptr)
    Store_field (modifiers, 1, hash_variant ("TFPtr"));
  else if (flags == Smoke::tf_ref)
    Store_field (modifiers, 1, hash_variant ("TFRef"));

  return modifiers;
}


